// created on 6/6/2004 at 10:46 AM
/*
 *   Copyright (c) 2004, Alexandros Frantzis (alf82 [at] freemail [dot] gr)
 *
 *   This file is part of Bless.
 *
 *   Bless is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Bless is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Bless; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
using NUnit.Framework;
using System.Collections;
using System;
using System.IO;

namespace Bless {

///<summary>Interface for ByteBuffer actions</summary>
interface IAction {
	void Do();
	void Undo();
}

///<summary>Action for appending at the end of a ByteBuffer</summary>
class AppendAction: IAction {

	Segment seg;
	ByteBuffer byteBuf;
	
	public AppendAction(byte[] d, ByteBuffer bb) {
		byteBuf=bb;
		ChangeBuffer cb=byteBuf.changeBuf;
		seg=new Segment(cb, cb.Size, cb.Size+d.Length-1);
		cb.Append(d);
		
	}
	
	public void Do() {
		// use copy of segment seg to protect it from alterations 
		byteBuf.segCol.Append(new Segment(seg.buffer, seg.start, seg.end));
		byteBuf.size+=seg.Size;
	}
	
	public void Undo() {
		byteBuf.size-=seg.Size;
		byteBuf.segCol.DeleteRange(byteBuf.size, byteBuf.size+seg.Size-1);
	}

}

///<summary>Action for inserting in a ByteBuffer</summary>
class InsertAction: IAction {

	Segment seg;
	long pos;
	ByteBuffer byteBuf;
	
	public InsertAction(long p, byte[] d, ByteBuffer bb) {
		byteBuf=bb;
		pos=p;
		ChangeBuffer cb=byteBuf.changeBuf;
		seg=new Segment(cb, cb.Size, cb.Size+d.Length-1);
		cb.Append(d);
	}
	
	public void Do() {
		SegmentCollection tmp=new SegmentCollection();
		// use copy of segment seg to protect it from alterations
		tmp.Append(new Segment(seg.buffer, seg.start, seg.end));
		byteBuf.segCol.Insert(tmp, pos);
		byteBuf.size+=seg.Size;
	}
	
	public void Undo() {
		byteBuf.segCol.DeleteRange(pos, pos+seg.Size-1);
		byteBuf.size-=seg.Size;
	}

}

///<summary>Action for deleting from a ByteBuffer</summary>
class DeleteAction: IAction {

	SegmentCollection del;
	long pos1, pos2;
	ByteBuffer byteBuf;
	
	public DeleteAction(long p1, long p2, ByteBuffer bb) {
		byteBuf=bb;
		pos1=p1;
		pos2=p2;
	}
	
	public void Do() {
		del=byteBuf.segCol.DeleteRange(pos1, pos2);
		byteBuf.size-=pos2-pos1+1;
	}
	
	public void Undo() {
		byteBuf.segCol.Insert(del, pos1);
		byteBuf.size+=pos2-pos1+1;
	}

}

class ReplaceAction: IAction {

	DeleteAction del;
	InsertAction ins;
	
	public ReplaceAction(long p1, long p2, byte[] d, ByteBuffer bb) {
		del=new DeleteAction(p1, p2, bb);
		ins=new InsertAction(p1, d, bb);
	}
	
	public void Do() {
		del.Do();
		ins.Do();
	}
	
	public void Undo() {
		ins.Undo();
		del.Undo();
	}

}

///<summary>
/// A buffer for holding bytes in a versatile manner.
/// It supports undo-redo and can easily handle large files.
/// Editing is also very cheap.
///</summary>
public class ByteBuffer {

	internal ChangeBuffer changeBuf;
	internal FileBuffer	fileBuf;
	internal SegmentCollection segCol;
	Stack undoStack;
	Stack redoStack;
	object SaveCheckpoint;
	
	internal long size; 
	
	public ByteBuffer() {
		segCol=new SegmentCollection();
		changeBuf= new ChangeBuffer();
		undoStack= new Stack();
		redoStack= new Stack();
		size=0;
		SaveCheckpoint=null;
	}
	
	public ByteBuffer(string filename): this() 
	{
		LoadWithFileBuffer(new FileBuffer(filename, 0xfffff));
		//Console.WriteLine("Loaded {0} {1}", filename, fileBuf.Size);
	}
	
	public delegate void ChangedHandler(); 
	
	public event ChangedHandler Changed;
	
	///<summary>Append bytes at the end of the buffer</summary>
	public void Append(byte[] data) 
	{
		AppendAction aa=new AppendAction(data, this);	
		aa.Do();
		undoStack.Push(aa);
		redoStack.Clear();
		if (Changed!=null)
			Changed();
	}

	///<summary>Insert bytes into the buffer</summary>
	public void Insert(long pos, byte[] data) 
	{
		if (pos == size) {
			Append(data);
			return;
		}
		InsertAction ia=new InsertAction(pos, data, this);
		ia.Do();		
		undoStack.Push(ia);
		redoStack.Clear();
		if (Changed!=null)
			Changed();
	}		

	///<summary>Delete bytes from the buffer</summary>
	public void Delete(long pos1, long pos2) 
	{
		DeleteAction da=new DeleteAction(pos1, pos2, this);
		da.Do();		
		undoStack.Push(da);
		redoStack.Clear();
		if (Changed!=null)
			Changed();
	}
	
	///<summary>Replace bytes in the buffer</summary>
	public void Replace(long pos1, long pos2, byte[] data) 
	{
		ReplaceAction ra=new ReplaceAction(pos1, pos2, data, this);
		ra.Do();		
		undoStack.Push(ra);
		redoStack.Clear();
		if (Changed!=null)
			Changed();
	}
	
	///<summary>Undo the last action</summary>
	public void Undo() 
	{
		if (undoStack.Count==0)
			return;
		IAction action=(IAction)undoStack.Pop();
		action.Undo();
		redoStack.Push(action);
		if (Changed!=null)
			Changed();
	}
	
	///<summary>Redo the last undone action</summary>
	public void Redo() 
	{
		if (redoStack.Count==0)
			return;
		IAction action=(IAction)redoStack.Pop();
		action.Do();
		undoStack.Push(action);
		if (Changed!=null)
			Changed();
	}
	
	public class SaveInterruptedException : Exception { }
	
	// a delegate that is called while saving.
	// 'fraction' is the amount of job done eg 0.45=45%
	// 'show' whether to show the dialog (it is hidden by default)
	// the delegate returns whether the save should be canceled (true)
	// or may continue (false)   
	public delegate bool SaveProgressDelegate(double fraction, bool show);
	
	//<summary>method to call when save timer expires</summary> 
	void SaveTimerExpired(object o)
	{
		// make the SaveProgressDialog visible
		SaveProgressDelegate deleg=(SaveProgressDelegate)o;
		deleg(0.0, true);
	}
	
	
	///<summary>Save the buffer as a file</summary>
	public void SaveAs(string filename, bool reload, SaveProgressDelegate deleg) 
	{
		// create 1 second timer
		System.Threading.Timer timer= new System.Threading.Timer(new System.Threading.TimerCallback(SaveTimerExpired), deleg, 1000, 0);
	
		FileStream fs=new FileStream(filename, FileMode.Create, FileAccess.Write);
		const int blockSize=0xffff;
		
		byte[] baTemp=new byte[blockSize];
		
		// for every node
		Util.List.Node node=segCol.List.First;
		bool cancel=false;
		
		while (node!=null && !cancel)
		{
			// Save the data in the node 
			// in blocks of blockSize each
			Segment s=(Segment)node.data;
			long len=s.Size;
			long nBlocks=len/blockSize;
			int last=(int)(len%blockSize); // bytes in last block
			long i;
			
			// for every full block
			for (i=0;i<nBlocks;i++) {
				s.buffer.Get(baTemp, s.start+i*blockSize, blockSize);
				fs.Write(baTemp, 0, blockSize);
				if (deleg!=null) { // update and decide if we should cancel
					cancel=deleg(((double)((i+1)*blockSize))/len, false);
					if (cancel) break;	
				}
			}
		
			// if last non-full block is not empty
			if (last!=0 && !cancel) {
				s.buffer.Get(baTemp, s.start+i*blockSize, last);
				fs.Write(baTemp, 0, last);
			}
			
			if (deleg!=null) // update and decide if we should cancel
				cancel=deleg(((double)(i*blockSize))/len, false);
					
			node=node.next;
		}
		
		fs.Close();
		timer.Dispose();
		
		// if save was canceled...
		if (cancel)
			throw new SaveInterruptedException();
		
		//Console.WriteLine("Saved to {0} {1}", filename, i);
		
		if (reload==true) {
			CloseFile();
			LoadWithFileBuffer(new FileBuffer(filename));
		}
		
		if (undoStack.Count > 0)
			SaveCheckpoint=undoStack.Peek();
		else
			SaveCheckpoint=null;
	}
	
	///<summary>Save the buffer under the same filename</summary>
	public void Save(SaveProgressDelegate deleg) 
	{
		string orig=this.Filename;
		if ((orig==null)||(orig==""))
			throw new System.ArgumentException("Filename is empty.");
			
		// 1. Save the buffer as a temporary file
		// 2. Delete the original file
		// 3. Move the temporary file to its proper position
		
		string tmp="/tmp/" + Util.TempFile.CreateName("/tmp");
		
		try {
			SaveAs(tmp, false, deleg);
			CloseFile();
			System.IO.File.Delete(orig); // this could throw
			System.IO.File.Move(tmp, orig);
			LoadWithFileBuffer(new FileBuffer(orig));
		}
		catch (SaveInterruptedException ex) {
			// delete temporary file
			System.IO.File.Delete(tmp);
			throw;
		}
		catch (Exception ex) // if something goes wrong use the old file
		{
			LoadWithFileBuffer(new FileBuffer(orig, 0xfffff));
			throw;	
		}
		
		if (undoStack.Count > 0)
			SaveCheckpoint=undoStack.Peek();
		else
			SaveCheckpoint=null;
	}
	
	///<summary> 
	/// Sets the file buffer and resets the segment collection
	///</summary> 
	private void LoadWithFileBuffer(FileBuffer fb)
	{
		fileBuf=fb;
		Segment s=new Segment(fileBuf, 0, fileBuf.Size-1);
		segCol=new SegmentCollection();
		segCol.Append(s);
		size=fileBuf.Size;
	}
	
	public byte this[long index] {
		set { } 
		get {
			long map; 
			Util.List.Node node;
			Segment seg=segCol.FindSegment(index, out map, out node);
			//Console.WriteLine("Searching index {0} at {1}:{2}", index, map, seg);
			if (seg==null)
				throw new IndexOutOfRangeException(string.Format("ByteBuffer[{0}]",index));
			else {
				try {
					return seg.buffer[seg.start+index-map];	
				}
				catch(IndexOutOfRangeException e) {
					Console.WriteLine("Problem at index {0} at {1}:{2}", index, map, seg);
					throw;
				}
			}
		}
	}
	
	
	public void CloseFile()
	{
		if (fileBuf != null)
				fileBuf.Close();
	} 
	
	public long Size {
		get { return size;}
	}

	public string Filename {
		get {
			if (fileBuf != null)
				return fileBuf.Filename;
			else
				return null; 
			}	
	}

	public bool HasChanged {
		get {
			if (undoStack.Count>0)
				return (SaveCheckpoint != undoStack.Peek());
			else
				return SaveCheckpoint != null;
		}
	
	}
	
	public bool CanUndo {
		get { return (undoStack.Count>0); }
	}
	
	public bool CanRedo {
		get { return (redoStack.Count>0); }
	}
	
	internal void Display(string s) 
	{
		Console.Write(s);
		segCol.List.Display();
	}

}


#region Tests
[TestFixture]
public class ByteBufferTests {

	[Test]
	public void AppendDoTest() {
		ByteBuffer bb=new ByteBuffer();
		byte[] ba={1,3,5,7,13,17};
		bb.Append(ba);
		
		Assert.AreEqual(6, bb.Size);
		Assert.AreEqual(1, bb[0]);
		Assert.AreEqual(3, bb[1]);
		Assert.AreEqual(5, bb[2]);
		Assert.AreEqual(7, bb[3]);
		Assert.AreEqual(13, bb[4]);
		Assert.AreEqual(17, bb[5]);
	}
	
	[Test]
	public void AppendUndoTest() {
		ByteBuffer bb=new ByteBuffer();
		byte[] ba={1,3,5,7};
		byte[] ba1={13,17};
		bb.Append(ba);
		bb.Append(ba1);
			
		Assert.AreEqual(6, bb.Size,"#1");
		Assert.AreEqual(3, bb[1],"#2");
		Assert.AreEqual(17, bb[5],"#3");
		
		bb.Undo();
	
		Assert.AreEqual(4, bb.Size,"#4");
		Assert.AreEqual(1, bb[0],"#5");
		Assert.AreEqual(3, bb[1],"#6");
		Assert.AreEqual(5, bb[2],"#7");
		Assert.AreEqual(7, bb[3],"#8");
	
	}
	
	[Test]
	public void InsertDoTest() {
		ByteBuffer bb=new ByteBuffer();
		byte[] ba={1,3,17};
		byte[] ba1={5,7,13};
		bb.Append(ba);
		//bb.Display("D1");
		bb.Insert(2, ba1);
		//bb.Display("D2");
		
		Assert.AreEqual(6, bb.Size,"#1");
		Assert.AreEqual(1, bb[0],"#2");
		Assert.AreEqual(3, bb[1],"#3");
		Assert.AreEqual(5, bb[2],"#4");
		Assert.AreEqual(7, bb[3],"#5");
		Assert.AreEqual(13, bb[4],"#6");
		Assert.AreEqual(17, bb[5],"#7");
	}
	
	[Test]
	public void InsertUndoTest() {
		ByteBuffer bb=new ByteBuffer();
		byte[] ba={1,3,17};
		byte[] ba1={5,7,13};
		bb.Append(ba);
		//bb.Display("D1");
		bb.Insert(2, ba1);
		//bb.Display("D2");
	
		Assert.AreEqual(6, bb.Size,"#1");
		Assert.AreEqual(3, bb[1],"#2");
		Assert.AreEqual(17, bb[5],"#3");
		
		bb.Undo();
		//bb.Display("D3");
		Assert.AreEqual(3, bb.Size,"#4");
		Assert.AreEqual(1, bb[0],"#5");
		Assert.AreEqual(3, bb[1],"#6");
		Assert.AreEqual(17, bb[2],"#7");
	}

	[Test]
	public void DeleteDoTest() {
		ByteBuffer bb=new ByteBuffer();
		byte[] ba={1,3,17};
		byte[] ba1={5,7,13};
		bb.Append(ba);
		//bb.Display("D1");
		bb.Insert(2, ba1);
		//bb.Display("D2");
		bb.Delete(1,3);
		
		Assert.AreEqual(3, bb.Size,"#1");
		Assert.AreEqual(1, bb[0],"#2");
		Assert.AreEqual(13, bb[1],"#3");
		Assert.AreEqual(17, bb[2],"#4");
	}
	
	[Test]
	public void DeleteUndoTest() {
		ByteBuffer bb=new ByteBuffer();
		byte[] ba={1,3,17};
		byte[] ba1={5,7,13};
		bb.Append(ba);
		//bb.Display("D1");
		bb.Insert(2, ba1);
		//bb.Display("D2");
		bb.Delete(1,3);
		
		Assert.AreEqual(3, bb.Size,"#1");
		Assert.AreEqual(1, bb[0],"#2");
		Assert.AreEqual(13, bb[1],"#3");
		Assert.AreEqual(17, bb[2],"#4");
		
		bb.Undo();
		
		Assert.AreEqual(6, bb.Size,"#5");
		Assert.AreEqual(1, bb[0],"#6");
		Assert.AreEqual(3, bb[1],"#7");
		Assert.AreEqual(5, bb[2],"#8");
		Assert.AreEqual(7, bb[3],"#9");
		Assert.AreEqual(13, bb[4],"#10");
		Assert.AreEqual(17, bb[5],"#11");
		
	}

	[Test]
	public void FileLoadTest1() {
		ByteBuffer bb=new ByteBuffer("test2.bin");
	
		long size=bb.Size;
		int sum=0;
		
		for(int i=0; i<size; i++) 
			sum ^=bb[i];
	
		Assert.AreEqual( 0x3a, sum);
		
	}
	
	[Test]
	public void FileLoadTest2() {
		ByteBuffer bb=new ByteBuffer("test2.bin");
		byte[] ba={0x41, 0x61};
		bb.Append(ba);
		
		long size=bb.Size;
		int sum=0;
		
		for(int i=0; i<size; i++) 
			sum ^=bb[i];
	
		Assert.AreEqual( 0x3a^0x20, sum);
		
	}
	
	[Test]
	public void FileLoadTest3() {
		ByteBuffer bb=new ByteBuffer("test2.bin");
		byte[] ba={0x41, 0x61};
		bb.Insert(10000, ba);
		
		long size=bb.Size;
		int sum=0;
		
		for(int i=0; i<size; i++) 
			sum ^=bb[i];
	
		Assert.AreEqual( 0x3a^0x20, sum);
		
	}
	
	[Test]
	public void FileSaveTest1() {
		ByteBuffer bb=new ByteBuffer("test2.bin");
		byte[] ba={0x41, 0x61};
		long size1=bb.Size;
		
		bb.Insert(10000, ba);
		Assert.AreEqual(size1+2, bb.Size,"#1");
		
		bb.SaveAs("test2.bak", false, null);
		
		ByteBuffer bb1=new ByteBuffer("test2.bak");
		
		
		Assert.AreEqual(bb.Size, bb1.Size,"#2");
		long size=bb1.Size;
		
		for(int i=0; i<size; i++) { 
			if (bb1[i]!=bb[i])
				Assert.Fail(string.Format("#3 Difference at {0} ({1}!={2})",i, bb1[i], bb[i]));
		}
		
	}

	[Test]
	public void HasChangedTest() 
	{
		ByteBuffer bb=new ByteBuffer();
		byte[] ba={0x41, 0x61};
		Assert.AreEqual(false, bb.HasChanged,"#0");
		bb.Append(ba);
		Assert.AreEqual(true, bb.HasChanged,"#1");
		bb.Undo();
		Assert.AreEqual(false, bb.HasChanged,"#2");
	}

}

#endregion

}