// created on 6/14/2004 at 10:35 PM
/*
 *   Copyright (c) 2004, Alexandros Frantzis (alf82 [at] freemail [dot] gr)
 *
 *   This file is part of Bless.
 *
 *   Bless is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Bless is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Bless; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
using System.Collections;
using System;
using Gtk;
using Bless.Gui.Areas;
using Bless.Gui.Drawers;
using Bless.Util;
namespace Bless.Gui {

 
///<summary>A widget that displays data from a buffer</summary>
public class DataView : Gtk.HBox {
	Layout layout;
	ByteBuffer byteBuffer;
	Gtk.DrawingArea drawingArea;
	Gtk.VScrollbar vscroll;
	bool widgetRealized;
	
	Gtk.Statusbar offsetBar;
	Gtk.Statusbar selectionBar;
	Gtk.Statusbar overwriteBar;
	
	// mouse selection related
	long selStart;
	bool selStartInAbyss;
	
	// keyboard selection related
	bool shiftPressed;
	bool overwrite;
	
	byte[] clipboard;
	
	//undo-redo related
	Stack cursorUndoStack;
	Stack cursorRedoStack; 
	
	// Filter keypresses
	Gtk.IMContext imContext;
	
	// class to store cursor position for undo/redo
	private class CursorHistory {
		public long UndoOffset;
		public int UndoDigit;
		public long RedoOffset;
		public int RedoDigit;
		public CursorHistory(long uo, int ud, long ro, int rd) 
			{ UndoOffset=uo; RedoOffset=ro; UndoDigit=ud; RedoDigit=rd;}
	}

	
	public enum GotoType { Closest, Start, End, Cursor }
	
	public ByteBuffer Buffer {
		get { return byteBuffer; }
		set { 
			byteBuffer=value;
			byteBuffer.Changed+=OnByteBufferChanged;
			foreach(Area a in layout.Areas) {
				a.Buffer=byteBuffer;
				a.CursorOffset=0;
				a.CursorDigit=0;
				a.Selection.Clear();
			}
			Redraw();
			vscroll.Adjustment.Value=0; 
		}
	} 
	
	public Layout Layout {
		get { return layout; }
		
		set { 
			// temporarily save previous layout
			Layout prevLayout=layout;
			
			// dispose Area pixmaps
			foreach (Area a in prevLayout.Areas)
				a.DisposePixmaps();
			
			// set new layout
			layout=value;
			
			if (widgetRealized) {
				layout.Realize(drawingArea, byteBuffer);
				Gdk.Rectangle alloc=drawingArea.Allocation;
				Resize(alloc.Width, alloc.Height);

				long prevOffset=0;
				 
				// Setup new areas according to the old ones 
				if (prevLayout.Areas.Count>0) {
					Area prevArea0=prevLayout.Areas[0] as Area;
					prevOffset=prevArea0.CursorOffset;
					foreach(Area a in layout.Areas) {
						a.CursorOffset=prevOffset;
						a.Selection.Start=prevArea0.Selection.Start;
						a.Selection.End=prevArea0.Selection.End;
					}
				}
				else {
					foreach(Area a in layout.Areas) 
						a.CursorOffset=0;
				}
						
				// make sure cursor is visible		
				Goto(prevOffset, GotoType.Closest);			
			}
		}
		
	}
	
	public Gtk.Statusbar OffsetBar {
		set { offsetBar=value;}
	}
	
	public Gtk.Statusbar SelectionBar {
		set { selectionBar=value;}
	}
	
	public Gtk.Statusbar OverwriteBar {
		set { overwriteBar=value;}
	
	}
	
	public bool Overwrite {
		set { overwrite=value;}
		get { return overwrite; }
	}
	
	///<summary>Create a DataView</summary>
	public DataView()
	{
		// load the default layout from the resources
		layout = new Layout();
		System.Reflection.Assembly a=System.Reflection.Assembly.GetExecutingAssembly();
		System.IO.Stream stream=a.GetManifestResourceStream("default.layout");
		string l=(new System.IO.StreamReader(stream)).ReadToEnd();
		layout.LoadXml(l);
		
		// initialize clipboard buffer
		clipboard=new byte[0];
		
		// initialize undo/redo cursor offset stacks
		cursorUndoStack=new Stack();
		cursorRedoStack=new Stack();
		 
		// initialize scrollbar
		Gtk.Adjustment 
			adj=new Gtk.Adjustment(0.0, 0.0, 1.0,1.0, 10.0, 0.0);
		vscroll=new Gtk.VScrollbar(adj); 
		
		adj.ValueChanged+=OnScrolled;
		
		// initialize drawing area
		drawingArea=new Gtk.DrawingArea();
		drawingArea.Realized+=OnRealized;
		drawingArea.ExposeEvent+=OnExposed;
		drawingArea.ConfigureEvent+=OnConfigured;
		drawingArea.ButtonPressEvent+=OnButtonPress;
		drawingArea.ButtonReleaseEvent+=OnButtonRelease;
		drawingArea.MotionNotifyEvent+=OnMotionNotify;
		drawingArea.KeyPressEvent+=OnKeyPress;
		drawingArea.KeyReleaseEvent+=OnKeyRelease;
		drawingArea.ScrollEvent+=OnMouseWheel;
		drawingArea.ModifyBg(StateType.Normal, new Gdk.Color(0xff,0xff,0xff));
		
		// add events that we want to handle
		drawingArea.AddEvents((int)Gdk.EventMask.ButtonPressMask);
		drawingArea.AddEvents((int)Gdk.EventMask.ButtonReleaseMask);
		drawingArea.AddEvents((int)Gdk.EventMask.PointerMotionMask);
		drawingArea.AddEvents((int)Gdk.EventMask.PointerMotionHintMask);
		drawingArea.AddEvents((int)Gdk.EventMask.KeyPressMask);
		drawingArea.AddEvents((int)Gdk.EventMask.KeyReleaseMask);
		
		drawingArea.CanFocus=true; // needed to catch key events
		
		imContext=new IMContextSimple();
		
		this.PackStart(drawingArea , true, true, 0);
		this.PackStart(vscroll , false, false, 0);
	}
	
	///<summary>Force a complete redraw of the view</summary>
	public void Redraw()
	{
		if (!widgetRealized)
			return;
		
		Gdk.Rectangle alloc=drawingArea.Allocation;
		Resize(alloc.Width, alloc.Height);
		drawingArea.QueueDraw();
		UpdateStatusBars();
	}
	
	///<summary>Redraw the view if something has changed</summary>
	void RedrawIfChanged()
	{
		foreach(Area a in layout.Areas) {
			a.Render(false);
		}
	 	UpdateStatusBars();
	}
	
	private void UpdateStatusBars()
	{
		if (layout.Areas.Count<=0)
			return;
		
		long coffset=((Area)layout.Areas[0]).CursorOffset;
		if (offsetBar!=null) {
			offsetBar.Pop(0);
			string str=string.Format("Cursor Offset: 0x{0:x} / 0x{1:x}", coffset, byteBuffer.Size-1);
			offsetBar.Push(0, str);
		}
		Util.Range sel=new Util.Range(((Area)layout.Areas[0]).Selection);
		sel.Sort();
		if (selectionBar!=null) {
			selectionBar.Pop(0);
			string str;
			
			if (sel.IsEmpty()==true)
				str="Selection: None";
			else	
				str=string.Format("Selection: 0x{0:x} to 0x{1:x} (0x{2:x} bytes)", sel.Start, sel.End, sel.Size);
				
			selectionBar.Push(0, str);
		}
	}
	
	private void UpdateOverwriteStatusBar()
	{
		if (overwriteBar==null)
			return;
			
		overwriteBar.Pop(0);
		if (overwrite==true)
			overwriteBar.Push(0, " OVR");
		else
			overwriteBar.Push(0, " INS");
	}
	
	void OnByteBufferChanged()
	{
		//System.Console.WriteLine("ByteBuffer changed!");
		Redraw();
	}
	
	///<summary>
	/// Find the number of bytes per row in order 
	/// to best utilize the available space and
	/// keep all areas synchronized.
	///</summary>
	private int FindBestBpr(int width)
	{
		int n=0; // current bpr
		int bestBpr=0; // best bpr so far
		int swBest=0; // width of best bpr so far 
		
		// try all values for n, from 0 upwards, 
		// until the width for a given n exceeds
		// the available. 
		while(true) {
			int sw=0; // width width current bpr
			
			foreach(Area a in layout.Areas) {
				int w=a.CalcWidth(n);
				// if this number of bpr is not acceptable
				// by an area (eg breaks grouping) skip it
				if (w==-1) goto next;  
				sw+=w;
			}
		
			// break if available width is exceeded
			// or last width value equals current one
			if (sw>width || sw==swBest)
				break;
		
			// keep best bpr so far
			bestBpr=n;
			swBest=sw;
			
			next:
			n++;
		}
		
		return bestBpr;
	}
	
	///<summary>Benchmark the rendering</summary>
	private void Benchmark() 
	{
		System.DateTime t1=System.DateTime.Now;
		
		
		for (int i=0; i<100; i++) {
			foreach(Area a in layout.Areas) {
				a.Render(true);
			}
		}
		
		System.DateTime t2=System.DateTime.Now;
		Gdk.Rectangle rect=drawingArea.Allocation;
		Console.WriteLine("100 render screen ({0},{1}): {2} ms", rect.Width, rect.Height,(t2-t1).Milliseconds);
	
	}
	
	///<summary>Get the area that contains (X,Y)</summary>
	private Area GetAreaByXY(int x, int y)
	{
		Area clickArea=null;
		
		foreach(Area a in layout.Areas) { 
			if ((x >= a.X) && (x <= a.X + a.Width)) {
				clickArea=a;
				break;
			}
		}
		
		return clickArea;
	}
	
	private void SetupScrollbarRange()
	{
		if (layout.Areas.Count<=0)
			return;
		
		long bpr=((Area)layout.Areas[0]).BytesPerRow;
		long nrows=((byteBuffer.Size+1)/bpr); // +1 because of append cursor position
			
		if (nrows < vscroll.Adjustment.PageSize) {
			vscroll.Value=0;
			vscroll.Hide();	
		}
		else if ((byteBuffer.Size+1)%bpr==0) { 
			vscroll.SetRange(0, nrows);
			vscroll.Show();
		}
		else {
			vscroll.SetRange(0, nrows+1);
			vscroll.Show();
		}
	
	}
	
	///<summary>Handles window resizing</summary>
	private void Resize(int winWidth, int winHeight)
	{
		// find bytes per row...
		int bpr=FindBestBpr(winWidth);
				
		// configure areas
		int s=0;
		int fontHeight=winHeight;
		foreach(Area a in layout.Areas) {
			a.Height=winHeight;
			a.Width=a.CalcWidth(bpr);
			a.X=s;
			a.BytesPerRow=bpr;
			s+=a.Width;
			if (a.Font.Height<fontHeight) {
				fontHeight=a.Font.Height;
			}
		}

		// configure scrollbar
		vscroll.Adjustment.PageSize=(winHeight/fontHeight);
		vscroll.SetIncrements(3, vscroll.Adjustment.PageSize-1);
		
		// decide whether the scrollbar is visible
		// and its range 
		if (bpr==0) {
			// this can cause eternal loop!
			// because hiding the scrollbar changes the
			// area and causes a reconfigure which may
			// show it again, and so on
			//vscroll.Hide();
		}
		else
			SetupScrollbarRange();
		
	}
	
	///<summary>Handle the Configure Event</summary>
	void OnConfigured (object o, ConfigureEventArgs args)
	{
		if (widgetRealized==false)
			return;
			
		Gdk.EventConfigure conf=args.Event;
		
		Resize(conf.Width, conf.Height);
	}
	
	///<summary>Handle the Expose Event</summary>
	void OnExposed (object o, ExposeEventArgs args)
	{
		foreach(Area a in layout.Areas) {
			a.Render(true);
		}
		UpdateStatusBars();
		UpdateOverwriteStatusBar();
 	}
 	
 	///<summary>Handle the Realized Event</summary>
	void OnRealized (object o, EventArgs args)
	{
				
		// Create some default areas
		layout.Realize(drawingArea, byteBuffer);
    	widgetRealized=true;
    	drawingArea.GrabFocus();
    	
    	// now we can configure properly
		Gdk.Rectangle alloc=((Widget)o).Allocation; 
    	Resize(alloc.Width, alloc.Height);
    	
	}
	
	///<summary>Handle scrolling</summary>
	void OnScrolled (object o, EventArgs args)
	{
	
		int bpr=0;
		if (layout.Areas.Count>0)
			bpr=((Area)layout.Areas[0]).BytesPerRow;
			
		long offset=(long)vscroll.Adjustment.Value*bpr;

		foreach(Area a in layout.Areas) 
			a.Offset=offset;	
			
		RedrawIfChanged();
		
	}
	
	///<summary>Scroll the view so that offset is visible</summary>
	public void Goto(long offset, GotoType type)
	{
		if (layout.Areas.Count<=0)
			return;
		
		int	bpr=((Area)layout.Areas[0]).BytesPerRow;
		if (bpr==0)
			return ;
		
		long curOffset=((Area)layout.Areas[0]).Offset;
		int h=((Area)layout.Areas[0]).Height;
		Drawer font=((Area)layout.Areas[0]).Font;
		int nrows=h/font.Height;
		
		long curOffsetRow=curOffset/bpr;
		long curOffsetEndRow=curOffsetRow+nrows-1;
		long offsetRow=offset/bpr;
		
		
		
		if (type == GotoType.Closest) {
			// if already visible do nothing
			if (offsetRow >= curOffsetRow && offsetRow <= curOffsetEndRow)
				;
			else if (curOffsetRow > offsetRow)
				type = GotoType.Start;
			else if (curOffsetEndRow < offsetRow)
				type = GotoType.End;
		}
		
		if (type == GotoType.Start) {
			 vscroll.Value=offsetRow;
		}
		else if (type == GotoType.End) {
			if (offsetRow-nrows>=0)
				vscroll.Value=offsetRow-nrows+1;
			else
				vscroll.Value=0;
		}
		else if (type == GotoType.Cursor) {
			long prevCursorRow=(layout.Areas[0] as Area).PrevCursorOffset/bpr;
			int diff=(int)(prevCursorRow-curOffsetRow);
			vscroll.Value = offsetRow - diff; 
		}
		
		
		RedrawIfChanged();
	}
	
	///<summary>Sets up the selection while dragging the mouse</summary>
	void EvaluateSelectionDrag(long selEnd, bool abyss)
	{
		long newSelStart=selStart;
		long newSelEnd=selEnd;
		
		
			
		if (selEnd!=selStart ) {
			if (selStart==byteBuffer.Size && !selStartInAbyss)
				newSelStart--;
			// find if the selection end is at a greater or
			// lesser position than the start and adjust
			// the selection accordingly
			 if (selEnd > selStart && abyss)
				newSelEnd--;
			 else if (selEnd < selStart && selStartInAbyss)
				newSelStart--;
				
			if (newSelEnd==byteBuffer.Size)
				newSelEnd--;
				
			foreach(Area a in layout.Areas) {
				a.Selection.Start=newSelStart;
 				a.Selection.End=newSelEnd;
 			}	
 			if (selEnd >= selStart && !abyss && selEnd<byteBuffer.Size) {
				foreach(Area a in layout.Areas){
					a.CursorOffset=selEnd+1;
					a.CursorDigit=0;
				}
			}
		}
		else if (((abyss && !selStartInAbyss) || (!abyss && selStartInAbyss)) && selStart!=byteBuffer.Size) {
			foreach(Area a in layout.Areas) {
				a.Selection.Start=selStart;
 				a.Selection.End=selStart;
 			}
 			if (selEnd >= selStart && !abyss && selEnd<byteBuffer.Size) {
				foreach(Area a in layout.Areas) {
					a.CursorOffset=selEnd+1;
					a.CursorDigit=0;
				}
			}
		}
		else {
			foreach(Area a in layout.Areas) 
				a.Selection.Clear();
		}
		
		
				
	}
	
	///<summary>Handle mouse button presses</summary>
	void OnButtonPress (object o, ButtonPressEventArgs args)
	{
		Gdk.EventButton e=args.Event;
		Area clickArea=GetAreaByXY((int)e.X, (int)e.Y);
		
		if (clickArea!=null) {
			Area.GetOffsetFlags flags;
			int digit;
			long off=clickArea.GetOffsetByDisplayInfo((int)(e.X-clickArea.X), (int)(e.Y-clickArea.Y), out digit, out flags);
			//Console.WriteLine("Press in {0} area @ {1:x}.{2}", clickArea.Type,off, digit);
			
			if ((flags & Area.GetOffsetFlags.Abyss) == Area.GetOffsetFlags.Abyss)
				off++;
			
			if ((flags & Area.GetOffsetFlags.Eof) == Area.GetOffsetFlags.Eof)
				off=byteBuffer.Size;
			
				
			// update cursor in areas if keypress outside selection
			if (!clickArea.Selection.Contains(off)) {	
				foreach(Area a in layout.Areas) {
					a.CursorOffset=off;
					a.Selection.Clear();
					a.HasCursorFocus=false;
				}
				// set the focused area
				clickArea.HasCursorFocus=true;
				clickArea.CursorDigit=digit;
			}
			
			if ((flags & Area.GetOffsetFlags.Abyss) == Area.GetOffsetFlags.Abyss)
				selStartInAbyss=true;
			else
				selStartInAbyss=false;
				
			selStart=off;
			
			RedrawIfChanged();
		}						  					
		
	}
	
	///<summary>Handle mouse motion</summary>
	void OnMotionNotify(object o, MotionNotifyEventArgs args)
	{
		Gdk.EventMotion e=args.Event;
		Gtk.DrawingArea da=(Gtk.DrawingArea)o;
		int x,y;
		Gdk.ModifierType state;
		
		if (e.IsHint)
    		da.GdkWindow.GetPointer(out x, out y, out state);
		else {
			x = (int)e.X;
			y = (int)e.Y;
			state = e.State;
		}
		
		// if left mouse button is down
		if ((state & Gdk.ModifierType.Button1Mask) != 0) {
					
			// find in which area the pointer is
			Area clickArea=GetAreaByXY(x, y);
						
			if (clickArea!=null) {
				Area.GetOffsetFlags flags;
				int digit;
				long off=clickArea.GetOffsetByDisplayInfo((int)(x-clickArea.X), (int)(y-clickArea.Y), out digit, out flags);
				//Console.WriteLine("Motion in {0} area @ {1:x}", clickArea.Type,off);
				if ((flags & Area.GetOffsetFlags.Abyss) == Area.GetOffsetFlags.Abyss)
					off++;
				if ((flags & Area.GetOffsetFlags.Eof) == Area.GetOffsetFlags.Eof)
					off=byteBuffer.Size;
				if (off<0)
					off=0;
				
				// update cursor in areas	
				foreach(Area a in layout.Areas) {
					a.CursorOffset=off;
					a.HasCursorFocus=false;		
				}
			
				// set the focused area
				clickArea.HasCursorFocus=true;
				clickArea.CursorDigit=digit;
				
				bool abyss=((flags & Area.GetOffsetFlags.Abyss) == Area.GetOffsetFlags.Abyss);
				
				EvaluateSelectionDrag(off, abyss);
				Goto(off, GotoType.Closest);
			}						  					
				
		}// end: if mouse button down
		
	}
	
	///<summary>Handle mouse button release</summary>
	void OnButtonRelease (object o, ButtonReleaseEventArgs args)
	{
		Gdk.EventButton e=args.Event;
		Area clickArea=GetAreaByXY((int)e.X, (int)e.Y);
		
		
		if (clickArea!=null) {
			int x=(int)e.X;
			int y;
			// if the pointer has moved out of the area height borders
			// make sure there is consistency in the selection
			// when depressing the mouse button
			if (e.Y > clickArea.Height + clickArea.Y) 
				y=clickArea.Height + clickArea.Y - clickArea.Font.Height;
			else if (e.Y < clickArea.Y)
				y=clickArea.Y;
			else
				y=(int)e.Y;
				
			Area.GetOffsetFlags flags;
			int digit;
			long off=clickArea.GetOffsetByDisplayInfo((int)(x-clickArea.X), (int)(y-clickArea.Y), out digit, out flags);
			//Console.WriteLine("Release in {0} area @ {1:x}", clickArea.Type,off);
			if ((flags & Area.GetOffsetFlags.Abyss) == Area.GetOffsetFlags.Abyss)
				off++;
			if ((flags & Area.GetOffsetFlags.Eof) == Area.GetOffsetFlags.Eof)
				off=byteBuffer.Size;
			
			// update cursor in areas	
			foreach(Area a in layout.Areas) {
				a.CursorOffset=off;
				a.Selection.Clear();
				if (a!=clickArea)
					a.HasCursorFocus=false;
					
			}
			// set the focused area
			clickArea.HasCursorFocus=true;
			clickArea.CursorDigit=digit;
			
			bool abyss=((flags & Area.GetOffsetFlags.Abyss) == Area.GetOffsetFlags.Abyss);
				
			EvaluateSelectionDrag(off, abyss);	
			RedrawIfChanged();
		}						  					
	}
	
	///<summary>Handle key presses</summary>
	void OnKeyPress (object o, KeyPressEventArgs args)
	{
		Gdk.EventKey e=args.Event;
		Area focusArea=null;
		int faIndex;
		
		// find focused area
		for (faIndex=0; faIndex < layout.Areas.Count; faIndex++) {
			if ((layout.Areas[faIndex] as Area).HasCursorFocus==true) {
				focusArea=(layout.Areas[faIndex] as Area);
				break;
			}
		}
		
		// if no area has the focus, give it to the first one applicable
		if (focusArea == null)
		for (faIndex=0; faIndex < layout.Areas.Count; faIndex++) {
			Area a=(layout.Areas[faIndex] as Area);
			if (a.Type!="offset" && a.Type !="separator") {
				a.HasCursorFocus=true;
				a.CursorDigit=0;
				focusArea=a;
				break;
			}
		}
		
		// if still no area got the focus give up
		if (focusArea == null)
			return;
		
		int bpr=focusArea.BytesPerRow;
		int dpb=focusArea.DigitsPerByte;
		long cOffset=focusArea.CursorOffset;
		int cDigit=focusArea.CursorDigit;
		int pageIncrement=(int)vscroll.Adjustment.PageIncrement;
		long bbSize=byteBuffer.Size;
		long selEnd=-1;
		GotoType gotoType=GotoType.Closest;
		
		// 
		if ((e.State & Gdk.ModifierType.ShiftMask) == Gdk.ModifierType.ShiftMask) {
			if (shiftPressed==false) {
				selStart=cOffset;
				shiftPressed=true;
			}
		}
		
		// handle keys when shift is not pressed
		if (!shiftPressed) {
			bool specialKey=false;
			
			switch(e.Key) {
			case Gdk.Key.Up:
				cOffset -= bpr;
				if (cOffset < 0) {
					cOffset=focusArea.CursorOffset;
					cDigit=focusArea.CursorDigit;
				}
				break;
			case Gdk.Key.Down:
				cOffset += bpr;
				if (cOffset >= bbSize) {
					cOffset = bbSize;
					cDigit=focusArea.CursorDigit;
				}
				break;
			case Gdk.Key.Left:
				cDigit--;
				if (cDigit<0) {
					cOffset--;
					cDigit = dpb - 1;
				}
				if (cOffset < 0) {
					cOffset = 0;
					cDigit=0;
				}	
				break;
			case Gdk.Key.Right:
				cDigit++;
				if (cDigit>=dpb) {
					cOffset++;
					cDigit = 0;
				}
				if (cOffset > bbSize) {
					cOffset = bbSize;
					cDigit = dpb-1;
				}
				break;
			case Gdk.Key.Page_Up:
				cOffset -= bpr*pageIncrement;
				if (cOffset < 0) {
					cOffset=0;
					cDigit=focusArea.CursorDigit;
				}
				gotoType=GotoType.Cursor;
				break;
			case Gdk.Key.Page_Down:
				cOffset += bpr*pageIncrement;
				if (cOffset > bbSize) {
					cOffset = bbSize;
					cDigit=focusArea.CursorDigit;
				}
				gotoType=GotoType.Cursor;
				break;
			case Gdk.Key.Home:
				cOffset = 0;
				cDigit = 0;
				gotoType=GotoType.Start;
				break;
			case Gdk.Key.End:
				cOffset = bbSize;
				cDigit = 0;
				gotoType=GotoType.End;
				break;
			case Gdk.Key.Insert:
				overwrite = !overwrite;
				UpdateOverwriteStatusBar();
				break;
			case Gdk.Key.Tab: // cycle focus
				for (int i=faIndex+1; (i % layout.Areas.Count)!= faIndex ;i++) {
					Area a=(layout.Areas[i%layout.Areas.Count] as Area);
					if (a.Type!="offset" && a.Type !="separator") {
						a.HasCursorFocus=true;
						focusArea.HasCursorFocus=false;
						a.CursorDigit=0;
						break;
					}
				}
				specialKey=true;
				break;
			default: // any other key pass it to focused area
				// if area handled it move one position right
				if (imContext.FilterKeypress(e) && focusArea.HandleKey(e.Key, overwrite) == true) {
					cDigit++;
					// bytebuffer size may have changed (eg append)
					bbSize=byteBuffer.Size;
					if (cDigit>=dpb) {
						cOffset++;
						cDigit = 0;
						cursorUndoStack.Push(new CursorHistory(cOffset-1, dpb-1, cOffset, cDigit));
					}
					else 
						cursorUndoStack.Push(new CursorHistory(cOffset, cDigit-1, cOffset, cDigit));
						
					cursorRedoStack.Clear();	
					
					/*if (cOffset > bbSize) {
						cOffset = bbSize;
						cDigit = dpb-1;
					}*/
				}
				else
					specialKey=true;
				break;
			} // end switch()
			
			if (!specialKey)
			foreach(Area a in layout.Areas) 
				a.Selection.Clear();
			
		} // end if (!shiftPressed)
		
		
		// handle keys when shift is pressed
		if (shiftPressed) {
			switch(e.Key) {
			case Gdk.Key.Up:
				cOffset -= bpr;
				if (cOffset < 0) {
					cOffset=focusArea.CursorOffset;
					selEnd=-1;
					cDigit=focusArea.CursorDigit;
				}
				else {
					selEnd=cOffset;
					cDigit=0;
				}
				break;
			case Gdk.Key.Down:
				cOffset += bpr;
				if (cOffset > bbSize) {
					cOffset = bbSize;
					selEnd=bbSize;
					cDigit=focusArea.CursorDigit;
				}
				else {	
					selEnd=cOffset;
					cDigit=0;
				}
				break;
			case Gdk.Key.Left:
				cOffset--;
				if (cOffset < 0)
					cOffset = 0;
				else {
					selEnd=cOffset;	
					cDigit=0;
				}	
				break;
			case Gdk.Key.Right:
				cOffset++;
				if (cOffset > bbSize)
					cOffset = bbSize;
				else {
					selEnd=cOffset;	
					cDigit=0;
				}
				break;
			case Gdk.Key.Page_Up:
				cOffset -= bpr*pageIncrement;
				if (cOffset < 0) {
					cOffset=0;
					selEnd=0;
					cDigit=focusArea.CursorDigit;
				}
				else {
					selEnd=cOffset;
					cDigit=0;
				}
				gotoType=GotoType.Cursor;
				break;
			case Gdk.Key.Page_Down:
				cOffset += bpr*pageIncrement;
				if (cOffset > bbSize) {
					cOffset=bbSize;
					selEnd=bbSize;
					cDigit=focusArea.CursorDigit;
				}
				else {
					selEnd=cOffset;
					cDigit=0;
				}
				gotoType=GotoType.Cursor;
				break;
			case Gdk.Key.Home:
				cOffset = 0;
				cDigit = 0;
				selEnd = cOffset;
				gotoType=GotoType.Start;
				break;
			case Gdk.Key.End:
				cOffset = bbSize;
				cDigit = 0;
				selEnd = cOffset;
				gotoType=GotoType.End;
				break;	
			default: // any other key pass it to focused area
				// if area handled it move one position right
				if (imContext.FilterKeypress(e) && focusArea.HandleKey(e.Key, overwrite) == true) {
					cDigit++;
					// bytebuffer size may have changed (eg append)
					bbSize=byteBuffer.Size;
					if (cDigit>=dpb) {
						cOffset++;
						cDigit = 0;
						cursorUndoStack.Push(new CursorHistory(cOffset-1, dpb-1, cOffset, cDigit));
					}
					else 
						cursorUndoStack.Push(new CursorHistory(cOffset, cDigit-1, cOffset, cDigit));
						
					cursorRedoStack.Clear();
					
					foreach(Area a in layout.Areas)
						a.Selection.Clear();
				}
				break;		
			} // end switch
			
			// if we have a new selection end
			if (selEnd != -1) {
				long newSelEnd=selEnd;
				long newSelStart=selStart;
				
				if (selEnd!=selStart) {
					// find if the selection end is at a greater or
					// lesser position than the start and adjust
					// the selection accordingly
					if (selEnd > selStart)
						newSelEnd--;
					else if (selEnd < selStart)
						newSelStart--;
					foreach(Area a in layout.Areas) {
						a.Selection.Start=newSelStart;
 						a.Selection.End=newSelEnd;
 					}	
				}
				else {
					foreach(Area a in layout.Areas) 
						a.Selection.Clear();
				}
				
 			}		
		} // end if (shiftPressed)
		
		// if cursor position should change
		if (cDigit != focusArea.CursorDigit || cOffset != focusArea.CursorOffset) {
			focusArea.CursorDigit=cDigit;
			foreach(Area a in layout.Areas) 
				a.CursorOffset=cOffset;
			
			Goto(cOffset, gotoType);
		}	
	}
	
	///<summary>Handle key releases</summary>
	void OnKeyRelease (object o, KeyReleaseEventArgs args)
	{
		Gdk.EventKey e=args.Event;
		if ( (e.Key == Gdk.Key.Shift_L || e.Key == Gdk.Key.Shift_R) )
			shiftPressed=false;
	}
	
	void OnMouseWheel(object o, ScrollEventArgs args)
	{
		Gdk.EventScroll e=args.Event;
		
		if (e.Direction == Gdk.ScrollDirection.Down) {
			double newValue=vscroll.Value + vscroll.Adjustment.StepIncrement;
			if (newValue <= vscroll.Adjustment.Upper - vscroll.Adjustment.PageSize)
		 		vscroll.Value=newValue;
		 	else
		 		vscroll.Value=vscroll.Adjustment.Upper - vscroll.Adjustment.PageSize;
		} 	
		else if (e.Direction == Gdk.ScrollDirection.Up) { 
			double newValue=vscroll.Value - vscroll.Adjustment.StepIncrement;
			if (newValue >= vscroll.Adjustment.Lower)
		 		vscroll.Value=newValue;
		 	else
		 		vscroll.Value=vscroll.Adjustment.Lower;
		 }
	}
	
	public void Copy()
	{
		if (layout.Areas.Count<=0)
			return;
		
		Area area0=((Area)layout.Areas[0]);
		area0.Selection.Sort();
		long clipSize=area0.Selection.Size;
		
		// if no selection, do nothing (keep old clipboard data)
		if (clipSize==0)
			return;
			
		clipboard=new byte[clipSize];
		
		long i=0;
		long start=area0.Selection.Start;
		while (i < clipSize) {
			clipboard[i]=byteBuffer[start+i];
			i++;
		}
		Goto(area0.CursorOffset, GotoType.Closest);
	}
	
	public void Cut()
	{
		if (layout.Areas.Count<=0)
			return;
		
		Area area0=((Area)layout.Areas[0]);
		area0.Selection.Sort();
		long clipSize=area0.Selection.Size;
		
		// if no selection, do nothing (keep old clipboard data)
		if (clipSize==0)
			return;
		
		clipboard=new byte[clipSize];
		
		long i=0;
		long start=area0.Selection.Start;
		while (i < clipSize) {
			clipboard[i]=byteBuffer[start+i];
			i++;
		}
		
		byteBuffer.Delete(area0.Selection.Start, area0.Selection.End);
		cursorUndoStack.Push(new CursorHistory(area0.CursorOffset, 0, area0.Selection.Start, 0));
		cursorRedoStack.Clear();
		
		//set cursor		
		foreach (Area a in layout.Areas)
			a.CursorOffset = area0.Selection.Start;
			  
		//clear selections
		foreach(Area a in layout.Areas) {
			a.Selection.Clear();
		}
		
		// Make sure Redraw() is called before Goto()
		// so that the Scrollbar has the correct range
		// when calling Goto(). 
		// Redraw();
		Goto(area0.CursorOffset, GotoType.Closest);
	}
	
	public void Paste()
	{
		if (layout.Areas.Count<=0)
			return;
		
		Area area0=((Area)layout.Areas[0]);
		
		if (clipboard.Length == 0)
			return;
		
		// if no range is selected insert the data and move
		// cursor to the end of inserted data
		if (area0.Selection.IsEmpty()){
			byteBuffer.Insert(area0.CursorOffset, clipboard);
			cursorUndoStack.Push(new CursorHistory(area0.CursorOffset, 0, area0.CursorOffset+clipboard.Length, 0));
			cursorRedoStack.Clear();
			
			foreach (Area a in layout.Areas)
				a.CursorOffset += clipboard.Length;
		}	
		else {
			// if a range is selected, replace the range with
			// the data and move the cursor to the end of
			// the new data
			byteBuffer.Replace(area0.Selection.Start, area0.Selection.End, clipboard);
			cursorUndoStack.Push(new CursorHistory(area0.Selection.Start, 0, area0.Selection.Start+clipboard.Length, 0));
			cursorRedoStack.Clear();
			
			foreach (Area a in layout.Areas){
				a.CursorOffset = a.Selection.Start + clipboard.Length;  
				a.Selection.Clear();
			}
		}
		
		// Make sure Redraw() is called before Goto()
		// so that the Scrollbar has the correct range
		// when calling Goto().
		// Redraw();
		Goto(area0.CursorOffset, GotoType.Closest);
	}
	
	public void Delete()
	{
		if (layout.Areas.Count<=0)
			return;
		
		Area area0=((Area)layout.Areas[0]);
		
		// if nothing is selected, delete the byte at the current offset
		if (area0.Selection.IsEmpty()==true) {
			if (area0.CursorOffset < byteBuffer.Size) {
				byteBuffer.Delete(area0.CursorOffset, area0.CursorOffset);
				cursorUndoStack.Push(new CursorHistory(area0.CursorOffset, area0.CursorDigit, area0.CursorOffset, area0.CursorDigit));
				cursorRedoStack.Clear();
			}
		}
		else { // delete the selection 
			area0.Selection.Sort();
			byteBuffer.Delete(area0.Selection.Start, area0.Selection.End);
			cursorUndoStack.Push(new CursorHistory(area0.CursorOffset, 0, area0.Selection.Start, 0));
			cursorRedoStack.Clear();
			
			//long bbSize=byteBuffer.Size;
			long cOffset=area0.Selection.Start;
			//if (cOffset >= bbSize)
			//	cOffset = bbSize-1; */
			foreach (Area a in layout.Areas) {
				a.CursorOffset=cOffset;
				a.Selection.Clear();
			}
		}
		
		// Make sure Redraw() is called before Goto()
		// so that if the buffer has changed size,
		// the Scrollbar has the correct range when calling Goto().
		// Redraw();
		Goto(area0.CursorOffset, GotoType.Closest);		
	}
	
	public void Undo()
	{
		// undo buffer changes
		byteBuffer.Undo();
		
		// restore cursor position
		if (cursorUndoStack.Count>0) {
			CursorHistory ch=(CursorHistory)cursorUndoStack.Pop();
			cursorRedoStack.Push(ch);
			foreach (Area a in layout.Areas) {
				a.CursorOffset=ch.UndoOffset;
				a.CursorDigit=ch.UndoDigit;	
			}
			Goto(ch.UndoOffset, GotoType.Closest);
		} 
	}
	
	public void Redo()
	{
		// redo buffer changes
		byteBuffer.Redo();
		
		// restore cursor position
		if (cursorRedoStack.Count>0) {
			CursorHistory ch=(CursorHistory)cursorRedoStack.Pop();
			cursorUndoStack.Push(ch);
			foreach (Area a in layout.Areas) {
				a.CursorOffset=ch.RedoOffset;
				a.CursorDigit=ch.RedoDigit;
			}
			Goto(ch.UndoOffset, GotoType.Closest);
		} 
	}
	
}// end DataView

}//namespace